# Copyright 2020 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def measures_update(path=None, version=None, force=False, logger=None, auto_update_rules=False, use_astron_obs_table=False):
    """
    Update or install the IERS data used for measures calculations from the ASTRON server at path.
    
    Original data source used by ASTRON is here: https://www.iers.org/IERS/EN/DataProducts/data.html

    If no update is necessary then this function will silently return.

    CASA maintains a separate Observatories table which is available in the casarundata
    collection through pull_data and data_update. The Observatories table found at ASTRON
    is not installed by measures_update and any Observatories file at path will not be changed
    by using this function.

    A text file (readme.txt in the geodetic directory in path) records the measures version string
    and the date when that version was installed in path.

    If path is None then config.measurespath is used.

    If the version requested matches the one in that text file then this function does
    nothing unless force is True.

    If a specific version is not requested (the default) and the modification time of that text
    file is less than 24 hrs before now then this function does nothing unless force is True. When this
    function checks for a more recent version and finds that the installed version is the most recent
    then modification time of that text file is checked to the current time even though nothing has
    changed in path. This limits the number of attempts to update the measures data (including checking f\
    or more recent data) to once per day. When the force argument is True and a specific version is
    not requested then this function always checks for the latest version.

    When auto_update_rules is True then path must exist and contain the expected readme.txt file.
    Path must be owned by the user, force must be False, and the version must be None. This 
    option is used during casatools initialization when measures_auto_update is True. Automatic 
    updating happens during casatools initialization so that the updated measures are in place 
    before any tool needs to use them.

    Using measures_update after casatools has started should always be followed by exiting 
    and restarting casa (or the casatools module if modular casa components are being used).

    A file lock is used to prevent more that one data update (pull_data, measures_update,
    or data_update) from updating any files in path at the same time. When locked, the 
    lock file (data_update.lock in path) contains information about the process that
    has the lock. When measures_update gets the lock it checks the readme.txt file in path
    to make sure that an update is still necessary (if force is True then an update 
    always happens). If the lock file is not empty then a previous update of path (pull_data,
    data_update, or measures_update) did not exit as expected and the contents of path are
    suspect. In that case, an error will be reported and nothing will be updated. The lock
    file can be checked to see the details of when that file was locked. The lock file can be
    removed and measures_update can be tried again. It may be safest in that case to remove path
    completely or use a different path and use pull_data to install a fresh copy of the
    desired version.

    Care should be used when using measures_update outside of the normal automatic
    update that other casa sessions are not using the same measures at the same time,
    especially if they may also be starting at that time. If a specific version is
    requested or force is True there is a risk that the measures may be updated while
    one of those other sessions are trying to load the same measures data, leading to
    unpredictable results. The lock file will prevent simulateous updates from
    happening but if each simultanous update eventually updates the same measures
    location (because force is True or the updates are requesting different versions)
    then the measures that any of those simultanous casatools modules sees is 
    unpredictable. Avoid multiple, simultanous updates outside of the automatic
    update process.

    **Note:** during auto updates, measures_update requires that the expected 
    readme.txt file already exists in the geodetic directory at path. If that file does 
    not exist or can not be interpreted as expected then measures_update will 
    return without updating any data.

    **Note:** if auto_update_rules is True the user must own path (in addition to having 
    read and write permissions there). The version must then also be None and the force option 
    must be False.

    **Note::** During use outside of auto updates, if path does not exist it will be created
    by this function.

    **Notes::** During use outside of auto updates, if the readme.txt file exists but can not
    be read as expected **OR** that file does not exist but the contents of path appear to
    contain measures data (table names in the expected locations) then this function will
    print messages describing that and exit without changing anything at path. Using
    a force value of True will disable this check and install measures at path even if path
    is not empty or the readme.txt file can not be read. This use of force should be used
    with caution.


    Parameters
       - path (str=None) - Folder path to place updated measures data. Must contain a valid geodetic/readme.txt. If not set then config.measurespath is used.
       - version (str=None) - Version of measures data to retrieve (usually in the form of yyyymmdd-160001.ztar, see measures_available()). Default None retrieves the latest.
       - force (bool=False) - If True, always re-download the measures data. Default False will not download measures data if updated within the past day unless the version parameter is specified and different from what was last downloaded.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Default None writes messages to the terminal
       - auto_update_rules (bool=False) - If True then the user must be the owner of path, version must be None, and force must be False.
        
    Returns
       None
    
    """
    import os
    import pkg_resources
    from datetime import datetime
    import sys

    from ftplib import FTP
    import tarfile
    import re
    import ssl
    import urllib.request
    import certifi
    import fcntl

    from .print_log_messages import print_log_messages
    from .get_data_lock import get_data_lock
    from .get_data_info import get_data_info
    from .measures_available import measures_available
    
    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        print_log_messages('path is None and has not been set in config.measurespath. Provide a valid path and retry.', logger, True)
        return

    path = os.path.expanduser(path)

    if auto_update_rules:
        if version is not None:
            print_log_messages('auto_update_rules requires that version be None', logger, True)
            return
        if force:
            print_log_messages('force must be False when auto_update_rules is True', logger, True)
            return
        if (not os.path.isdir(path)) or (os.stat(path).st_uid != os.getuid()):
            msgs = []
            msgs.append("Warning: path must exist as a directory and it must be owned by the user, path = %s" % path)
            msgs.append("Warning: no measures auto update is possible on this path by this user.")
            print_log_messages(msgs, logger, False)
            return
    
    if not os.path.exists(path):
        # make dirs all the way down, if possible
        os.makedirs(path)
        
    current = None
    ageRecent = False

    # first, does this look like it needs to be updated

    # get any existing measures readme information
    readmeInfo = get_data_info(path, logger, type='measures')
    if readmeInfo is not None:
        current = readmeInfo['version']
        if readmeInfo['age'] is not None:
            ageRecent = readmeInfo['age'] < 1.0

    if not force:
        # don't check for new version if the age is less than 1 day
        if version is None and ageRecent:
            # normal use is silent, this line is useful during debugging
            # print_log_messages('measures_update latest version checked recently in %s, using version %s' % (path, current), logger)
            return
        
        # don't overwrite something that looks bad unless forced to do so
        if current == 'invalid':
            print_log_messages('The measures readme.txt file could not be read as expected, an update can not proceed unless force is True', logger, True)
            return

        # don't overwrite something that looks like valid measures data unless forced to do so
        if current == 'unknown':
            print_log_messages('The measures data at %s is not maintained by casaconfig and so it can not be updated unless force is True' % path, logger, True)
            return

        checkVersion = version
        if checkVersion is None:
            # get the current most recent version
            try:
                checkVersion = measures_available()[-1]
            except:
                # unsure what happened, leave it at none, which will trigger an update attempt, which might work
                pass

        # don't re-download the same data
        if (checkVersion is not None) and (checkVersion == current):
            # normal use is silent, this line is useful during debugging
            # print_log_messages('measures_update requested version already installed in %s' % path, logger)
            # update the age of the readme to now
            try:
                readme_path = os.path.join(path,'geodetic/readme.txt')
                # readme_path should already exist if it's here
                os.utime(readme_path)
            except:
                # unsure what happened, everything otherwise is fine if we got here, ignore this error
                pass
            
            return

        # don't do anything unless the Observatories table is already installed as expected
        obsTabPath = os.path.join(path,'geodetic/Observatories')
        if not os.path.isdir(obsTabPath):
            msgs = []
            msgs.append("Error: the Observatories table was not found as expected in %s" % path)
            msgs.append("Either install casarundata first or set use_astron_obs_table and force to be True when using measures_update.")
            msgs.append("Note that the Observatories table provided in the Astron measures tarfile is not the same as that maintained by CASA")
            print_log_messages(msgs, logger, True)
            return
        
    # path must be writable with execute bit set
    if (not os.access(path, os.W_OK | os.X_OK)) :
        print_log_messages('No permission to write to the measures path, cannot update : %s' % path, logger, True)
        return

    # an update needs to happen

    # lock the measures_update.lock file
    lock_fd = None
    try:
        print_log_messages('measures_update ... acquiring the lock ... ', logger)

        lock_fd = get_data_lock(path, 'measures_update')
        # if lock_fd is None it means the lock file was not empty - because we know that path exists at this point
        if lock_fd is None:
            # using a list of messages results in a better printing if the logger is redirected to the terminal
            msgs = []
            msgs.append('The lock file at %s is not empty.' % path)
            msgs.append('A previous attempt to update path may have failed or exited prematurely.')
            msgs.append('Remove the lock file and set force to True with the desired version (default to most recent).')
            msgs.append('It may be best to completely repopulate path using pull_data and measures_update.')
            print_log_messages(msgs, logger, True)
            return

        do_update = force

        if not do_update:
            # recheck the readme file, another update may have already happened before the lock was obtained
            current = None
            ageRecent = False
            
            readmeInfo = get_data_info(path, logger, type='measures')
            if readmeInfo is not None:
                current = readmeInfo['version']
                if readmeInfo['age'] is not None:
                    ageRecent = readmeInfo['age'] < 1.0

            if (version is not None) and (version == current):
                # no update will be done, version is as requested - not silent here because the lock is in use
                print_log_messages('The requested measures version is already installed in %s, using version %s' % (path, current), logger)
            elif (version is None) and ageRecent:
                # no update will be done, it's already been checked or updated recently - not silent here because the lock is in use
                print_log_messages('The latest measures version was checked recently in %s, using version %s' % (path, current), logger)
            else:
                # final check for problems before updating
                if not force and readmeInfo is not None and (version=='invalid' or version=='unknown'):
                    # at this point, this indicates something is unexpectedly wrong, do not continue
                    # using a list of messages results in a better printing if the logger is redirected to the terminal
                    msgs = []
                    msgs.append('Something unexpected has changed in the measures path location, and measures_update can not continue')
                    msgs.append('A previous measures_update may have exited unexpectedly')
                    msgs.append('It may be necessary to reinstall the casarundata as well as the measures data if %s is the correct path' % path)
                    print_log_messages(msgs, logger, True)
                    # update is already turned off, the lock file will be cleaned up on exit
                else:
                    # an update is needed
                    do_update = True

        if do_update:
            if force:
                print_log_messages('A measures update has been requested by the force argument', logger)

            print_log_messages('  ... connecting to ftp.astron.nl ...', logger)

            ftp = FTP('ftp.astron.nl')
            rc = ftp.login()
            rc = ftp.cwd('outgoing/Measures')
            files = sorted([ff for ff in ftp.nlst() if (len(ff) > 0) and (not ff.endswith('.dat')) and (ftp.size(ff) > 0)])

            # target filename to download
            # for the non-force unspecified version case this can only get here if the age is > 1 day so there should be a newer version
            # but that isn't checked - this could install a version that's already installed
            target = files[-1] if version is None else version
            if target not in files:
                print_log_messages('measures_update cant find specified version %s' % target, logger, True)

            else:
                # there are files to extract, make sure there's no past measures.ztar from a failed previous install
                ztarPath = os.path.join(path,'measures.ztar')
                if os.path.exists(ztarPath):
                    os.remove(ztarPath)
    
                with open(ztarPath, 'wb') as fid:
                    print_log_messages('  ... downloading %s from ASTRON server to %s ...' % (target, path), logger)
                    ftp.retrbinary('RETR ' + target, fid.write)

                ftp.close()

                # remove any existing measures readme.txt now in case something goes wrong during extraction
                readme_path = os.path.join(path,'geodetic/readme.txt')
                if os.path.exists(readme_path):
                    os.remove(readme_path)

                # extract from the fetched tarfile
                with tarfile.open(ztarPath, mode='r:gz') as ztar:
                    # the list of members to extract
                    x_list = []
                    for m in ztar.getmembers() :
                        if force and use_astron_obs_table:
                            # exclude the *.old names in geodetic
                           if not(re.search('geodetic',m.name) and re.search('.old',m.name)):
                                x_list.append(m)
                        else:
                            # exclude the Observatories table and  *.old names in geodetic
                            if not((re.search('geodetic',m.name) and re.search('.old',m.name)) or re.search('Observatories',m.name)):
                                x_list.append(m)

                    # use the 'data' filter if available, revert to previous 'fully_trusted' behavior of not available
                    ztar.extraction_filter = getattr(tarfile, 'data_filter', (lambda member, path: member))
                    ztar.extractall(path=path,members=x_list)
                    ztar.close()

                os.remove(ztarPath)

                # create a new readme.txt file
                with open(readme_path,'w') as fid:
                    fid.write("# measures data populated by casaconfig\nversion : %s\ndate : %s" % (target, datetime.today().strftime('%Y-%m-%d')))

                print_log_messages('  ... measures data update at %s' % path, logger)
            
            # closing out the do_update

        # closing out the try block
        # truncate the lock file
        lock_fd.truncate(0)
        
    except Exception as exc:
        msgs = []
        msgs.append("ERROR! : Unexpected exception while updating measures at %s" % path)
        msgs.append("ERROR! : %s" % exc)
        print_log_messages(msgs, logger, True)
        # leave the contents of the lock file as is to aid in debugging
        
    # if the lock file is not closed, do that now to release the lock
    if lock_fd is not None and not lock_fd.closed:
        lock_fd.close()

    return



# Copyright 2020 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""


def set_casacore_path(path=None):
    """
    Set the path in .casarc to the desired data directory

    The "measures.directory" in the .casarc file is not used by any part of CASA. This function is
    provided as a convenience for casacore users (e.g. python-casacore) so that the path to
    the casacore data directory can be found by casacore users. It sets the value of that
    parameter to the casacore data location, replacing any previously set value.

    Parameters
       - path (string=None) - path to the desired data directory. Default None uses the included data directory from this package

    Returns
       None

    """
    import pkg_resources
    import os
    import re
    import sys
    
    if path is None: path = pkg_resources.resource_filename('casaconfig', 'data/')
    path = os.path.abspath(os.path.expanduser(path))

    rctext = 'measures.directory: %s\n' % path

    casarc = os.path.expanduser('~/.casarc')
    
    # get the current .casarc contents if it exists
    # set/replace the path to the geodetic folder
    if os.path.exists(casarc):
        with open(casarc, 'r') as fid:
            print('found existing .casarc...', file = sys.stderr )
            oldtxt = fid.read()
        if 'measures.directory' in oldtxt:
            rctext = re.sub('measures.directory: .+?\n', '%s'%rctext, oldtxt, flags=re.DOTALL)
        else:
            rctext = oldtxt.strip() + '\n' + rctext
    # write out new .casarc file
    with open(casarc, 'w') as fid:
        print('writing %s...' % casarc)
        fid.write(rctext)
        
    return



# Copyright 2023 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def do_auto_updates(configDict, logger=None):
    """
    Use measurespath, data_auto_update, and measures_auto_update from configDict to
    do any auto updates as necessary.

    This is intended for use during casatools init but may be useful in other cases.

    Note that the IERS measures data is usually read when casatools starts and so 
    changes made to that table may not be seen until a new session if this function
    is used outside of casatools initialization.

    measurespath must be set (not None).

    measures_auto_update must be True when data_auto_update is True.

    See the documentation for data_update and measures_update for additional details
    about the auto update rules.

    Paramters
       - configDict (dict) - A config dictionary previously set. 
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Default None writes messages to the terminal.

    Returns
       None

    """

    from .print_log_messages import print_log_messages
    from .data_update import data_update
    from .measures_update import measures_update

    if configDict.measurespath is None:
        # continue, because things still might work if there are measures in datapath
        msgs = []
        msgs.append('measurespath is None in config')
        msgs.append('set measurespath in your config file at ~/.casa/config.py')
        msgs.append('or ask the site manager to set that in a casasiteconfig.py')
        msgs.append('visit https://casadocs.readthedocs.io/en/stable/notebooks/external-data.html for more information')

        if (configDict.measures_auto_update or configDict.data_auto_update):
            msgs.append('Auto updates of measures path are not possible because measurespath is not set, skipping auto updates')

        print_log_messages(msgs, logger, True)

        return

    if (configDict.measures_auto_update or configDict.data_auto_update):
        if (configDict.data_auto_update and (not configDict.measures_auto_update)):
            print_log_messages('measures_auto_update must be True when data_auto_update is True, skipping auto updates', logger, True)
        else:
            if configDict.data_auto_update:
                data_update(configDict.measurespath, logger=logger, auto_update_rules=True)
            if configDict.data_auto_update or configDict.measures_auto_update:
                measures_update(configDict.measurespath, logger=logger, auto_update_rules=True)

    return



# Copyright 2023 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def update_all(path=None, logger=None, force=False):
    """
    Update the data contants at path to the most recently released versions
    of casarundata and measures data. 

    If path does not exist it will be created (the user must have permission
    to create path).

    If path does exist and is not empty it must contain a previously
    installed version of casarundata. Path must be a directory and it
    must be owned by the user.

    If path is not provided then config is imported and the measurespath value set
    by that process will be used.

    If path already contains the most recent versions of casarundata and 
    measurespath then nothing will change at path.

    The force argument is passed to data_update and measures_update

    This uses pull_data, data_update and measures_update. See the
    documentation for those functions for additional details.

    Some of the data updated by this function is only read when casatools starts.
    Use of update_all after CASA has started should typically be followed by a restart 
    so that any changes are seen by the tools and tasks that use this data.

    Parameters
       - path (str=None) - Folder path to place casarundata contents. It must not exist, or be empty, or contain a valid, previously installed version. If it exists, it must be owned by the user. Default None uses the value of measurespath set by importing config.py.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Messages are always written to the terminal. Default None does not write any messages to a logger.

    Returns
       None

    """

    import os

    from .print_log_messages import print_log_messages
    from .get_data_info import get_data_info
    from .pull_data import pull_data
    from .data_update import data_update
    from .measures_update import measures_update

    if path is None:
        from casaconfig import config
        path = config.measurespath

        if path is None:
            print_log_messages("config.measurespath is None. Edit your config.py to set measurespath to an appropriate location,", logger, True)
            return

    if not os.path.exists(path):
        # create it, all the way down
        try:
            os.makedirs(path, exist_ok=True)
        except:
            print_log_messages("unable to create path, check the permissions of the directory it is being created at, path = %s" % path, logger, True)
            return

    # path must be a directory and it must be owned by the user
    
    if (not os.path.isdir(path)) or (os.stat(path).st_uid != os.getuid()):
        msgs = []
        msgs.append("Warning: path must exist as a directory and it must be owned by the user, path = %s" % path)
        msgs.append("Warning: no updates are possible on this path by this user.")
        print_log_messages(msgs, logger, False)
        return

    # if path is empty, first use pull_data
    if len(os.listdir(path))==0:
        pull_data(path, logger)
        # double check that it's not empty
        if len(os.listdir(path))==0:
            print_log_messages("pull_data failed, see the error messages for more details. update_all can not continue")
            return

    # readme.txt must exist in path at this point, using get_data_info provides more possible feedback
    dataInfo = get_data_info(path, logger)
    if dataInfo is None:
        # this should not happen
        print_log_messages('could not find any data information about %s, can not continue with update_all' % path, logger, True)

    if dataInfo['casarundata'] is None:
        print_log_messages('readme.txt not found at path, update_all can not continue, path = %s' % path, logger)
        return

    if dataInfo['casarundata'] is 'invalid':
        print_log_messages('readme.txt is invalid at path, update_all can not continue, path = %s' % path, logger)
        return

    if dataInfo['casarundata'] is 'unknown':
        print_log_messages('contents at path appear to be casarundata but no readme.txt was found, casaconfig did not populate this data and update_all can not continue, path = %s', path, logger)
        return

    # the updates should work now
    data_update(path, logger, force=force)
    measures_update(path, logger, force=force)

    return



# Copyright 2020 AUI, Inc. Washington DC, USA
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""
this module will be included in the api
"""

def pull_data(path=None, version=None, force=False, logger=None):
    """
    Pull the casarundata contents from the CASA host and install it in path.

    The path must either contain a previously installed version of casarundata 
    or it must not exist or be empty.

    If path is None then config.measurespath will be used.

    If version is None (the default) then the most recent version is pulled.

    If version is "release" then the version associated with the release in 
    the dictionary returned by get_data_info is used. If there is no release
    version information known then an error message is printed and nothing is
    checked or installed.

    If force is True then the requested version is installed even if that
    version is already installed.

    Results and errors are always printed. They are also logged to the logger
    if available.

    A text file (readme.txt at path) records the version string, the date
    when that version was installed in path, and the files installed into path.
    This file is used to determine if the contents are a previously installed 
    version. If path is not empty then this file must exist with the expected
    contents in order for pull_data to proceed.

    If the version to be pulled matches the version in the readme.txt file then 
    pull_data does nothing unless force is True. No error messages will result when the
    version already matches what was previously installed (no installation is then
    necessary unless force is True).

    The measures tables included in casarundata will typically not be the most 
    recent version. To get the most recent measures data, measures_update
    should be used after pull_data.

    If path contains a previously installed version then all of the files listed in 
    the manifest part of the readme.txt file are first removed from path. This ensures
    that files not present in the version being installed are removed in moving to the
    other version.

    A file lock is used to prevent more than one data update (pull_data, measures_update,
    or data_update) from updating any files in path at the same time. When locked, the
    lock file (data_update.lock in path) contains information about the process that
    has the lock. When pull_data gets the lock it checks the readme.txt file in path
    to make sure that a copy of the data should still be pulled (the version doesn't 
    match what was requested, or force is True). If the lock file is not
    empty then a previous update of path (pull_data, data_update, or measures_update)
    did not exit as expected and the contents of path are suspect. In that case, pull_data
    will report that as an error and nothing will be updated. The lock file can be checked
    to see the details of when that file was locked. The lock file can be removed and
    pull_data can then be used to install the desired version. It may be safest in that case
    to remove path completely or use a different path and run pull_data to install 
    a fresh copy of the desired version.

    Some of the tables installed by pull_data are only read when casatools starts. Use of 
    pull_data should typically be followed by a restart of CASA so that 
    any changes are seen by the tools and tasks that use this data.

    **Note:** When version is None (the default), data_available is always used to find out
    what versions are available. There is no check on when the data were last updated before
    calling data_available (as there is in the two update functions). 

    Parameters
       - path (str) - Folder path to place casarundata contents. It must be empty or not exist or contain a valid, previously installed version. If not set then config.measurespath is used.
       - version (str=None) - casadata version to retrieve. Default None gets the most recent version.
       - force (bool=False) - If True, re-download and install the data even when the requested version matches what is already installed. Default False will not download data if the installed version matches the requested version.
       - logger (casatools.logsink=None) - Instance of the casalogger to use for writing messages. Messages are always written to the terminal. Default None does not write any messages to a logger.

    Returns
       None

    """

    import os

    from .data_available import data_available
    from .print_log_messages import print_log_messages
    from .get_data_lock import get_data_lock
    from .do_pull_data import do_pull_data
    from .get_data_info import get_data_info

    if path is None:
        from .. import config as _config
        path = _config.measurespath

    if path is None:
        print_log_messages('path is None and has not been set in config.measurespath. Provide a valid path and retry.', logger, True)
        return

    # when a specific version is requested then the measures readme.txt that is part of that version
    # will get a timestamp of now so that default measures updates won't happen for a day unless the
    # force argument is used for measures_update
    namedVersion = version is not None

    path = os.path.expanduser(path)
    readme_path = os.path.join(path, 'readme.txt')

    installed_files = []
    available_data = None
    currentVersion = None
    currentDate = None
    
    # attempt a pull if path does not exist or is empty (except for any lock file, handled later)
    readmeInfo = get_data_info(path, logger, type='casarundata')
    do_pull = readmeInfo is None
    if not do_pull:
        # find the current version, install date, and installed files
        currentVersion = readmeInfo['version']
        currentDate = readmeInfo['date']
        installed_files = readmeInfo['manifest']

        if currentVersion == 'invalid':
            msgs = []
            msgs.append('destination path is not empty and this does not appear to be casarundata OR the readme.txt file found there could not be read as expected')
            msgs.append('choose a different path or empty this path and try again')
            print_log_messages(msgs, logger, True)
            # no lock has been set yet, safe to simply return here
            return

        if currentVersion == 'unknown':
            msgs = []
            msgs.append('destination path appears to be casarundata but no readme.txt file was found')
            msgs.append('no data will be installed but CASA use of this data may be OK. Choose a different path or delete this path to install new casarundata.')
            if force:
                msgs.append('force is True but there is no readme.txt found and the location is not empty, no data will be installed')
                msgs.append('Choose a different path or empty this path to install new casarundata')
            print_log_messages(msgs, logger, True)
            # no lock as been set yet, safe to simply return here
            return

        if (installed_files is None or len(installed_files) == 0):
            # this shouldn't happen
            msgs = []
            msgs.append('destination path is not empty and the readme.txt file found there did not contain the expected list of installed files')
            msgs.append('choose a different path or empty this path and try again')
            print_log_messages(msgs, logger, True)
            # no lock as been set yet, safe to simply return here
            return
            
        # the readme file looks as expected, pull if the version is different or force is true
        if version is None:
            # use most recent available
            available_data = data_available()
            version = available_data[-1]

        do_pull = (version!=currentVersion) or force

        if not do_pull:
            # it's already at the expected version and force is False, nothing to do
            # safe to return here, no lock has been set
            return
        
    # a pull will happen, unless the version string is not available

    if available_data is None:
        available_data = data_available()

    if version is None:
        # need a version, use most recent available
        version = available_data[-1]

    expectedMeasuresVersion = None
    if version == 'release':
        # use the release version from get_data_info
        releaseInfo = get_data_info()['release']
        if releaseInfo is None:
            print_log_messages('No release info found, pull_data can not continue', logger, True)
            return
        version = releaseInfo['casarundata']
        expectedMeasuresVersion = releaseInfo['measures']

    if version not in available_data:
        print_log_messages('version %s not found on CASA server, use casaconfig.data_available() for a list of casarundata versions' % version, logger, True)
        return

    if not os.path.exists(path):
        # make dirs all the way down path if possible
        os.makedirs(path)

    # lock the data_update.lock file
    lock_fd = None
    try:
        print_log_messages('pull_data using version %s, acquiring the lock ... ' % version, logger)

        lock_fd = get_data_lock(path, 'pull_data')
        # if lock_fd is None it means the lock file was not empty - because we know that path exists at this point
        if lock_fd is None:
            msgs = []
            msgs.append('The lock file at %s is not empty.' % path)
            msgs.append('A previous attempt to update path may have failed or exited prematurely.')
            msgs.append('Remove the lock file and set force to True with the desired version (default to the most recent).')
            msgs.append('It may be best to clean out that location and do a fresh pull_data.')
            print_log_messages(msgs, logger, True)
            return
        
        do_pull = True
        if not force:
            # need to recheck any readme file that may be present here
            # it's possible that another process had path locked and updated the readme file with new information
            readmeInfo = get_data_info(path, logger, type='casarundata')
            if readmeInfo is not None:
                currentVersion = readmeInfo['version']
                currentDate = readmeInfo['date']
                if ((currentVersion == version) and (not force)):
                    if expectedMeasuresVersion is not None:
                        # this is a release pull and the measures version must also match
                        # start off assuming a pull is necessary
                        do_pull = True
                        measuresReadmeInfo = get_data_info(path, logger, type='measures')
                        if measuresReadmeInfo is not None:
                            measuresVersion = measuresReadmeInfo['version']
                            if measuresVersion == expectedMeasuresVersion:
                                # no pull is necessary
                                do_pull = False
                                print_log_messages('pull_data requested "release" versions of casarundata and measures are already installed.', logger)
                            # otherwise this is a release pull and even if the measuresVersion is 'unknown' or 'invalid' we should proceed with a pull at this point
                            # problems with casarundata path will have been caught before
                    else:
                        # nothing to do here, already at the expected casarundata version and a pull is not being forced and this is not a release pull
                        do_pull = False
                        print_log_messages('pull_data requested version is already installed.', logger)

                # a version of 'invalid', 'error', or 'unknown' is a surprise here, likely caused by something else doing something
                # incompatible with this attempt
                if version in ['invalid','error','unknown']:
                    do_pull = False
                    msgs = []
                    msgs.append('Unexpected version or problem found in readme.txt file during pull_data, can not safely pull the requested version')
                    msgs.append('This should not happen unless multiple sessions are trying to pull_data at the same time and one experienced problems or was done out of sequence')
                    print_log_messages(msgs, logger, True)
                    

                if do_pull:
                    # make sure the copy of installed_files is the correct one
                    installed_files = readmeInfo['manifest']
                    if len(installed_files) == 0:
                        # this shoudn't happen, do not do a pull
                        do_pull = False
                        msgs = []
                        msgs.append('destination path is not empty and the readme.txt file found there did not contain the expected list of installed files')
                        msgs.append('This should not happen unless multiple sessions are trying to pull_data at the same time and one experienced problems or was done out of sequence')
                        msgs.append('Check for other updates in process or choose a different path or clear out this path and try again')
                        print_log_messages(msgs, logger, True)

        if do_pull:
            do_pull_data(path, version, installed_files, currentVersion, currentDate, logger)
            if namedVersion:
                # a specific version has been requested, set the times on the measures readme.txt to now to avoid
                # a default update of the measures data without using the force argument
                measuresReadmePath = os.path.join(path,'geodetic/readme.txt')
                os.utime(measuresReadmePath)

                
        # truncate the lock filed
        lock_fd.truncate(0)
        
    except Exception as exc:
        msgs = []
        msgs.append('ERROR! : Unexpected exception while populating casarundata version %s to %s' % (version, path))
        msgs.append('ERROR! : %s' % exc)
        print_log_messages(msgs, logger, True)
        # leave the contents of the lock file as is to aid in debugging
        # import traceback
        # traceback.print_exc()

    # if the lock file is not closed, do that now to release the lock
    if lock_fd is not None and not lock_fd.closed:
        lock_fd.close()
        
    return
